//=============================================================================
// Attribute Boost by Variable
// WangRenJie
//=============================================================================

/*:
 * @plugindesc Allows a state to boost character's attributes by a percentage based on variable values. Optimized version by 修改者名字.
 * @param BoostVariableId
 * @desc The ID of the variable used for attribute boosts. Optimized version removed default value.
 * @type variable
 *
 * @help Attribute Boost by Variable
 * Version: 1.1
 *
 * ---------------------------
 * Plugin Commands
 * ---------------------------
 *
 * To set the boost percentage for the state 100 according to variable 105:
 *
 *   SetStateBoost 100 105
 *
 * The first parameter is state ID, the second parameter is variable ID.
 *
 * To remove the boost effect from all states:
 *
 *   ClearStateBoost
 *
 * ---------------------------
 * State Notes
 * ---------------------------
 *
 * You can use the following notetag in a state's Note field:
 *
 * <AttributeBoost: x>
 *
 * Replace x with a number to manually set the boost percentage for this state.
 * This value will override any boost values set by plugin commands.
 *
 */

(function() {

  var pluginParams = PluginManager.parameters('AttributeBoostByVariable');
  var boostVarId = parseInt(pluginParams['BoostVariableId']);

  // Parse state notetags to retrieve user-set boost values.
  var _DataManager_extractMetadata = DataManager.extractMetadata;
  DataManager.extractMetadata = function(data) {
    _DataManager_extractMetadata.call(this, data);
    if (data.meta.AttributeBoost) {
      data.boostPercentage = parseInt(data.meta.AttributeBoost);
    }
  };

  // Parse plugin commands.
  var _Game_Interpreter_pluginCommand =
    Game_Interpreter.prototype.pluginCommand;
  Game_Interpreter.prototype.pluginCommand = function(command, args) {
    _Game_Interpreter_pluginCommand.call(this, command, args);
    if (command === 'SetStateBoost') {
      var stateId = parseInt(args[0]);
      var varId = parseInt(args[1]);
      $dataStates[stateId].boostPercentageByVar = varId;
    } else if (command === 'ClearStateBoost') {
      $dataStates.forEach(function(state) {
        state.boostPercentageByVar = undefined;
      });
    }
  };

  // Calculate attribute boost percentage based on the variable value.
  Game_BattlerBase.prototype.getAttributeBoostPercentage = function() {
    var boost = (boostVarId && $gameVariables.value(boostVarId)) || 0;
    return (boost > 0 ? boost : 0) / 100;
  };

  // Override the original methods to apply attribute boosts.
  var _Game_BattlerBase_param = Game_BattlerBase.prototype.param;
  Game_BattlerBase.prototype.param = function(paramId) {
    var value = _Game_BattlerBase_param.call(this, paramId);
    var boostPercentage = 0;
    var states = this.states();
    for (var i = 0; i < states.length; i++) {
      var state = states[i];
      if (state.boostPercentage) {
        boostPercentage += state.boostPercentage;
      } else if (state.boostPercentageByVar) {
        var varId = state.boostPercentageByVar;
        var boost = (varId && $gameVariables.value(varId)) || 0;
        boostPercentage += (boost > 0 ? boost : 0) / 100;
      }
    }
    boostPercentage += this.getAttributeBoostPercentage();
    return Math.round(value * (1 + boostPercentage));
  };

})();