//=============================================================================
// JHMW_Message.js
//=============================================================================
 /*:
 * @plugindesc v1.1 自定义消息弹出
 * @author 江湖梦网 Fanzi
 *
 * @param 默认位置
 * @desc 默认弹出位置，对照小键盘数字
 * @type number
 * @min 1
 * @max 9
 * @default 1
 *
 * @param 窗口透明度
 * @desc 默认窗口透明度
 * @type number
 * @min 0
 * @max 255
 * @default 128
 *
 * @param 背景颜色
 * @desc 通知的默认背景颜色，按窗口皮肤的颜色序号
 * @type number
 * @min 0
 * @default 15
 *
 * @param 显示时长
 * @desc 显示消息的持续时长(帧数60为一秒)
 * @type number
 * @min 0
 * @default 120
 *
 * @param 过度时长
 * @desc 显示与消退过度时长(帧数60为一秒)
 * @type number
 * @min 0
 * @default 40
 *
 * @param 获得文本
 * @desc 获得金钱、物品、经验的显示内容
 * @default 获得了
 *
 * @param 失去文本
 * @desc 失去金钱、物品、经验的显示内容
 * @default 失去了
 *
 * @param 升级文本
 * @desc 角色等级提高时的显示内容
 * @default 升到了
 *
 * @param 降级文本
 * @desc 角色等级降低时的显示内容
 * @default 降到了
 *
 * @param 金钱位置
 * @desc 获得金钱的弹出位置
 * @type number
 * @min 1
 * @max 9
 * @default 1
 *
 * @param 金钱文本
 * @desc 获得金钱的弹出格式，留空不弹出
 * %1 - 得失文本    %2 - 金额    %3 - 单位
 * @default %1\i[314]%2%3
 *
 * @param 经验位置
 * @desc 获得经验的弹出位置
 * @type number
 * @min 1
 * @max 9
 * @default 1
 *
 * @param 经验文本
 * @desc 获得经验的弹出格式，留空不弹出
 * %1 - 得失文本    %2 - 角色名    %3 - 经验值
 * @default %2%1%3经验值
 *
 * @param 物品位置
 * @desc 获得物品的弹出位置
 * @type number
 * @min 1
 * @max 9
 * @default 1
 *
 * @param 物品文本
 * @desc 获得物品的弹出格式，留空不弹出
 * %1 - 得失文本    %2 - 数量    %3 - 物品名(含图标)
 * @default %1%2份%3
 *
 * @param 升级位置
 * @desc 升级降级的弹出位置
 * @type number
 * @min 1
 * @max 9
 * @default 1
 *
 * @param 等级文本
 * @desc 等级提升的弹出格式，留空不弹出
 * %1 - 升降文本    %2 - 角色名    %3 - 级数
 * @default %2%1%3级
 *
 * @help
 * 这是一个高级自定义输出消息的辅助插件，输出消息时不会暂停游戏，
 * 也不用等待按键确认，消息自动显示自动隐退，自动关联输出与否。
 * 你可以设定事件指令或者战斗掉落的金钱、道具、经验、升级等等是
 * 否自动显示，也可以自定义相应的显示内容，你要采用人工输出也行。
 * 共创建了九宫格输出与自定义坐标输出，全场景皆可输出。
 *
 * 使用这些数字来确定通知的位置，参照小键盘。
 * 7   8   9
 * 4   5   6
 * 1   2   3
 * 如果使用 0，需带坐标参数，通知窗口将出现在指定的坐标位置。
 *
 * 显示通知可以通过插件命令来实现，插件指令注意都有空格。如：
 * Notify: color x y text
 *   - 'color' 为通知的背景颜色
 *   - 'x' 为要显示的位置横坐标
 *   - 'y' 为要显示的位置纵坐标
 *   - 'text' 为要显示的信息内容，可以使用显示文章的所有代码
 * 示例：Notify: 0 200 200 太棒了！\i[123]
 *
 * 这将出现在插件参数设置的“默认位置”中。
 *   Notify1: text
 *   Notify2: text
 *   Notify3: text
 *   Notify4: text
 *   Notify5: text
 *   Notify6: text
 *   Notify7: text
 *   Notify8: text
 *   Notify9: text
 * 使用与数字板相对应的数字来确定屏幕位置。
 *
 * 需要变更通知的背景颜色可在前面使用颜色指令：
 *   NotifyColor: 2
 * 一个颜色指令只更改接下来要显示的一条通知的背景颜色。
 *
 * 显示通知也可以通过调用脚本来实现，用法基本类似。
 *   Notification.add(text, location, color, x, y);
 *   - 'text' 为要显示的信息内容，必填！可以使用显示文章的所有代码
 *   - 注意使用显示文章的代码时，脚本指令要比插件指令多一条反斜杠
 *   - 'location' 为显示位置，必填！若为九宫格位置后续参数可省略
 *   - 若为自定义坐标需带全参数，'x'、'y' 分别为显示坐标
 *   - 'color' 为背景颜色，位置填 0 时必填！按窗口皮肤字体颜色填数字
 *   - 'x' 为要显示的位置横坐标
 *   - 'y' 为要显示的位置纵坐标
 * 示例：Notification.add('获得\\i[314]！', 0, 1, 100, 100);
 * 或者：var text = '获得\\i[314]！';
 *        Notification.add(text, 1);
 *
 * 倒计时显示文章：
 * $gameMessage.add(text, true);
 * 比如在地图执行脚本：$gameMessage.add('检测到数据异常，为您重启游戏。', true);
 * 将用显示文章方式倒计时十秒返回标题界面，也可点击鼠标立即跳转标题界面。
 */
//=============================================================================
var JHMW = JHMW || {};
var Imported = Imported || {};
Imported.JHMW_Message = 1.1;
JHMW.Message = JHMW.Message || {};
JHMW.MessageParam = PluginManager.parameters('JHMW_Message');
JHMW.Message.DefaultLocation = Number(JHMW.MessageParam["默认位置"]);
JHMW.Message.WindowOpacity = Number(JHMW.MessageParam["窗口透明度"] || 255);
JHMW.Message.DefaultColor = Number(JHMW.MessageParam['背景颜色']);
JHMW.Message.DisplayFrames = Number(JHMW.MessageParam["显示时长"] || 120);
JHMW.Message.ToggleFrames = Number(JHMW.MessageParam["过度时长"] || 40);
JHMW.Message.GainText = String(JHMW.MessageParam['获得文本']);
JHMW.Message.LoseText = String(JHMW.MessageParam['失去文本']);
JHMW.Message.LevelUpText = String(JHMW.MessageParam['升级文本']);
JHMW.Message.LvDownText = String(JHMW.MessageParam['降级文本']);
JHMW.Message.MoneyLocation = Number(JHMW.MessageParam["金钱位置"]);
JHMW.Message.MoneyText = String(JHMW.MessageParam['金钱文本']);
JHMW.Message.LevelLocation = Number(JHMW.MessageParam["升级位置"]);
JHMW.Message.LevelText = String(JHMW.MessageParam['等级文本']);
JHMW.Message.ItemLocation = Number(JHMW.MessageParam["物品位置"]);
JHMW.Message.ItemText = String(JHMW.MessageParam['物品文本']);
JHMW.Message.ExpLocation = Number(JHMW.MessageParam["经验位置"]);
JHMW.Message.ExpText = String(JHMW.MessageParam['经验文本']);

function Notification() {
    throw new Error("This is a static class");
}

Notification = function() {};
Notification.windows0 = [];
Notification.windows1 = [];
Notification.windows2 = [];
Notification.windows3 = [];
Notification.windows4 = [];
Notification.windows5 = [];
Notification.windows6 = [];
Notification.windows7 = [];
Notification.windows8 = [];
Notification.windows9 = [];
Notification.nextColor = JHMW.Message.DefaultColor;

Notification.add = function(text, location, color, x, y) {
    if (location == undefined) location = JHMW.Message.DefaultLocation;
    if (color == undefined) color = Notification.nextColor;
    Notification.nextColor = JHMW.Message.DefaultColor;
    var array = this["windows" + location];
    var notificationWindow = new Window_MsgPopup(array.length, text, location, color, x, y);
    array.push(notificationWindow);
    SceneManager._scene.addChild(notificationWindow);
    if (!!location) Notification.refreshY(array);
    else Notification.refreshXY(array);
};

Notification.remove = function(location) {
    location = location || 0;
    location = location.clamp(0, 9);
    var array = this["windows" + location];
    var notificationWindow = array.shift();
    SceneManager._scene._windowLayer.removeChild(notificationWindow);
    if (location > 0) Notification.refreshY(array);
};

Notification.refreshY = function(array) {
    for (var i = 0; i < array.length; i++) {
        if (array[i]) array[i].y = array[i].calculateYPosition(i);
    }
};

Notification.refreshXY = function(array) {
    for (var i = 0; i < array.length; i++) {
        if (array[i]) {
            array[i].x = array[i].calculateYPosition(i).x;
            array[i].y = array[i].calculateYPosition(i).y;
        }
    }
};

Notification.removeAll = function() {
    for (var i = 0; i <= 9; i++) {
        var array = this["windows" + i];
        while(array.length != 0) {
            var notificationWindow = array.shift();
            SceneManager._scene._windowLayer.removeChild(notificationWindow);
        }
    }
};

JHMW.Message.Game_Message_clear = Game_Message.prototype.clear;
Game_Message.prototype.clear = function() {
    JHMW.Message.Game_Message_clear.call(this);
    this._popup = false;
};

Game_Message.prototype.add = function(text, type) {
    this._texts.push(text);
    if (type) this._popup = true;
    else this._popup = false;
};

Game_Message.prototype.isPopup = function() {
    return this._popup;
};

JHMW.Message.Game_Party_initialize = Game_Party.prototype.initialize;
Game_Party.prototype.initialize = function() {
    JHMW.Message.Game_Party_initialize.call(this);
    this._showPopups = false;
};

JHMW.Message.Game_Party_gainGold = Game_Party.prototype.gainGold;
Game_Party.prototype.gainGold = function(amount) {
    JHMW.Message.Game_Party_gainGold.call(this, amount);
    if (!JHMW.Message.MoneyText || amount == 0) return;
    if (amount > 0) var gainText = JHMW.Message.GainText;
    else var gainText = JHMW.Message.LoseText;
    var text = JHMW.Message.MoneyText.format(gainText, amount, TextManager.currencyUnit);
    var location = JHMW.Message.MoneyLocation;
    var color = Notification.nextColor;
    Notification.add(text, location, color, 0, 0);
};

JHMW.Message.Game_Party_gainItem = Game_Party.prototype.gainItem;
Game_Party.prototype.gainItem = function(item, amount, includeEquip) {
    JHMW.Message.Game_Party_gainItem.call(this, item, amount, includeEquip);
    if (!JHMW.Message.ItemText || this._showPopups || !item || !item.name || item.itypeId > 2 || amount == 0) return;
    if (amount > 0) var gainText = JHMW.Message.GainText;
    else var gainText = JHMW.Message.LoseText;
    if (item.textColor) var text = JHMW.Message.ItemText.format(gainText, amount, '\\c[' + item.textColor + ']\\i[' + item.iconIndex + ']' + item.name);
    else var text = JHMW.Message.ItemText.format(gainText, amount, '\\i[' + item.iconIndex + ']' + item.name);
    var location = JHMW.Message.ItemLocation;
    var color = Notification.nextColor;
    Notification.add(text, location, color, 0, 0);
};

JHMW.Message.Game_Actor_changeExp = Game_Actor.prototype.changeExp;
Game_Actor.prototype.changeExp = function(exp, show) {
    var tempexp = this._exp[this._classId];
    JHMW.Message.Game_Actor_changeExp.call(this, exp, show);
    if (!JHMW.Message.ExpText || exp - tempexp == 0) return;
    if (exp > tempexp) var gainText = JHMW.Message.GainText;
    else var gainText = JHMW.Message.LoseText;
    var text = JHMW.Message.ExpText.format(gainText, this._name, exp - tempexp);
    var location = JHMW.Message.ExpLocation;
    var color = Notification.nextColor;
    Notification.add(text, location, color, 0, 0);
};

JHMW.Message.Game_Actor_levelUp = Game_Actor.prototype.levelUp;
Game_Actor.prototype.levelUp = function() {
    JHMW.Message.Game_Actor_levelUp.call(this);
    if (!JHMW.Message.LevelText) return;
    var text = JHMW.Message.LevelText.format(JHMW.Message.LevelUpText, this._name, this._level);
    var location = JHMW.Message.LevelLocation;
    var color = Notification.nextColor;
    Notification.add(text, location, color, 0, 0);
};

JHMW.Message.Game_Actor_levelDown = Game_Actor.prototype.levelDown;
Game_Actor.prototype.levelDown = function() {
    JHMW.Message.Game_Actor_levelDown.call(this);
    if (!JHMW.Message.LevelText) return;
    var text = JHMW.Message.LevelText.format(JHMW.Message.LvDownText, this._name, this._level);
    var location = JHMW.Message.LevelLocation;
    var color = Notification.nextColor;
    Notification.add(text, location, color, 0, 0);
};

JHMW.Message.Game_Actor_changeEquip = Game_Actor.prototype.changeEquip;
Game_Actor.prototype.changeEquip = function(slotId, item) {
    $gameParty._showPopups = true;
    JHMW.Message.Game_Actor_changeEquip.call(this, slotId, item);
    $gameParty._showPopups = false;
};

JHMW.Message.Game_Actor_forceChangeEquip = Game_Actor.prototype.forceChangeEquip;
Game_Actor.prototype.forceChangeEquip = function(slotId, item) {
    $gameParty._showPopups = true;
    JHMW.Message.Game_Actor_forceChangeEquip.call(this, slotId, item);
    $gameParty._showPopups = false;
};

JHMW.Message.Game_Actor_tradeItemWithParty = Game_Actor.prototype.tradeItemWithParty;
Game_Actor.prototype.tradeItemWithParty = function(newItem, oldItem) {
    $gameParty._showPopups = true;
    var bool = JHMW.Message.Game_Actor_tradeItemWithParty.call(this, newItem, oldItem);
    $gameParty._showPopups = false;
    return bool;
};

if (Imported.YEP_ItemCore === true) {
    JHMW.Message.Game_Actor_initIndependentEquips = Game_Actor.prototype.initIndependentEquips;
    Game_Actor.prototype.initIndependentEquips = function(equips) {
        $gameParty._showPopups = true;
        JHMW.Message.Game_Actor_initIndependentEquips.call(this, equips);
        $gameParty._showPopups = false;
    };

    JHMW.Message.Game_Actor_changeEquipById = Game_Actor.prototype.changeEquipById;
    Game_Actor.prototype.changeEquipById = function(etypeId, itemId) {
        $gameParty._showPopups = true;
        JHMW.Message.Game_Actor_changeEquipById.call(this, etypeId, itemId);
        $gameParty._showPopups = false;
    };
}

Game_Interpreter.prototype.command316 = function() {
    var value = this.operateValue(this._params[2], this._params[3], this._params[4]);
    this.iterateActorEx(this._params[0], this._params[1], function(actor) {
        actor.changeLevel(actor.level + value, this._params[5]);
        if (JHMW.Message.LevelText && value != 0) {
            if (value > 0) var levelText = JHMW.Message.LevelUpText;
            else var levelText =JHMW.Message.LvDownText;
            var text = JHMW.Message.LevelText.format(levelText, actor._name, 'value');
            var location = JHMW.Message.LevelLocation;
            var color = Notification.nextColor;
            Notification.add(text, location, color, 0, 0);
        }
    }.bind(this));
    return true;
};

JHMW.Message.Game_Interpreter_command356 = Game_Interpreter.prototype.command356;
Game_Interpreter.prototype.command356 = function() {
    if (this._params[0].match(/Notify:/i) || this._params[0].match(/Notify(\d+):/i)) return this.commandArisuNotify();
    if (this._params[0].match(/NotifyColor:/i)) return this.commandArisuNotifyColor();
    return JHMW.Message.Game_Interpreter_command356.call(this);
};

Game_Interpreter.prototype.commandArisuNotify = function() {
    if (this._params[0].match(/Notify: (\d+) (\d+) (\d+)[ ](.*)/i)) {
        var location = 0;
        var color = Number(RegExp.$1);
        var x = Number(RegExp.$2);
        var y = Number(RegExp.$3);
        var text = String(RegExp.$4);
    } else if (this._params[0].match(/Notify(\d+):[ ](.*)/i)) {
        var location = Number(RegExp.$1);
        var color = Notification.nextColor;
        var x = 0;
        var y = 0;
        var text = String(RegExp.$2);
    } else if (this._params[0].match(/Notify:[ ](.*)/i)) {
        var location = JHMW.Message.DefaultLocation.clamp(1, 9);
        var color = Notification.nextColor;
        var x = 0;
        var y = 0;
        var text = String(RegExp.$1);
    }
    Notification.add(text, location, color, x, y);
    return true;
};

Game_Interpreter.prototype.commandArisuNotifyColor = function() {
    if (this._params[0].match(/NotifyColor: (\d+)(.*)/i)) {
        Notification.nextColor = Number(RegExp.$1);
    }
    return true;
};

JHMW.Message.Scene_Base_terminate = Scene_Base.prototype.terminate;
Scene_Base.prototype.terminate = function() {
    JHMW.Message.Scene_Base_terminate.call(this);
    Notification.removeAll();
};

Window_Message.prototype.textWidthEx = function(text) {
    return this.drawTextEx(text, 0, this.contents.height);
};

Window_Message.prototype.setPopup = function() {
    this.width = this.textWidthEx(this._textState.text + '我') + this.standardPadding() * 2;
    this.x = (Graphics.boxWidth - this.width) / 2;
    this.y = (Graphics.boxHeight - this.height) / 2;
    this.height = 72;
    this._countDown = 10;
    this._frameDown = 60;
    this._lastCount = 10;
};

var Window_Message_updatePlacement = Window_Message.prototype.updatePlacement;
Window_Message.prototype.updatePlacement = function() {
    this.width = this.windowWidth();
    this.height = this.windowHeight();
    this.x = (Graphics.boxWidth - this.width) / 2;
    Window_Message_updatePlacement.call(this);
};

var Window_Message_startMessage = Window_Message.prototype.startMessage;
Window_Message.prototype.startMessage = function() {
    Window_Message_startMessage.call(this);
    if ($gameMessage.isPopup()) this.setPopup();
};

Window_Message.prototype.refresh = function() {
    this.contents.fillRect(this.width - 64, 0, this.textWidthEx('我'), 36, this.textColor(7));
    this.drawText(this._countDown, this.width - 64, 0, this.textWidthEx('我'), 'center');
};

Window_Message.prototype.updateWait = function() {
    if (this._waitCount > 0) {
        this._waitCount--;
        return true;
    } else {
        if ($gameMessage.isPopup()) {
            this._frameDown--;
            if (this._frameDown <= 0) {
                this._countDown--;
                this._frameDown = 60;
            }
        }
        return false;
    }
};

Window_Message.prototype.update = function() {
    this.checkToNotClose();
    Window_Base.prototype.update.call(this);
    if (this._lastCount != this._countDown) {
        this.refresh();
        this._lastCount = this._countDown;
    }
    while (!this.isOpening() && !this.isClosing()) {
        if (this.updateWait()) {
            return;
        } else if (this.updateLoading()) {
            return;
        } else if (this.updateInput()) {
            if ($gameMessage.isPopup() && this._countDown <= 0) {
                //this.pause = false;
                this.terminateMessage();
            } else return;
        } else if (this.updateMessage()) {
            return;
        } else if (this.canStart()) {
            this.startMessage();
        } else {
            this.startInput();
            return;
        }
    }
};

var Window_Message_terminateMessage = Window_Message.prototype.terminateMessage;
Window_Message.prototype.terminateMessage = function() {
    if ($gameMessage.isPopup()) SceneManager.goto(Scene_Title);
    Window_Message_terminateMessage.call(this);
};

function Window_MsgPopup() {
    this.initialize.apply(this,arguments);
}

Window_MsgPopup.prototype = Object.create(Window_Base.prototype);
Window_MsgPopup.prototype.constructor = Window_MsgPopup;

Window_MsgPopup.prototype.initialize = function(index, text, location, color, x, y) {
    this.location = location;
    this.locationX = x || 0;
    this.locationY = y || 0;
    if (location > 0) {
        y = this.calculateYPosition(index);
        Window_Base.prototype.initialize.call(this, 0, y, Graphics.width, this.lineHeight());
    } else Window_Base.prototype.initialize.call(this, x, y, Graphics.width, this.lineHeight());
    this.loadWindowskin();
    this.bgcolor = this.textColor(color);
    this.frameCount = 0;
    this.isFadingIn = true;
    this.isFadingOut = false;
    this.isShowing = false;
    this.text = text;
    this.opacity = 0;
    this.contentsOpacity = 0;
    this.createBackground();
    this.refresh();
};

Window_MsgPopup.prototype.calculateYPosition = function(index) {
    switch(this.location) {
        case 0:
            var size = {};
            size.x = this.locationX;
            size.y = this.locationY;
            return size;
            break;
        case 1:case 2:case 3:
            var size = (Notification["windows" + this.location].length + 1) * this.lineHeight();
            return Graphics.boxHeight - size + this.lineHeight() * index;
            break;
        case 4:case 5:case 6:
            var size = Notification["windows" + this.location].length * this.lineHeight();
            return (Graphics.boxHeight - size) / 2 + this.lineHeight() * index;
            break;
        case 7:case 8:case 9:
            return this.lineHeight() * (index + 1);
            break;
    }
};

Window_MsgPopup.prototype.standardPadding = function() {
    return 0;
};

Window_MsgPopup.prototype.createBackground = function() {
    this._backSprite = new Sprite;
    this._backSprite.bitmap = new Bitmap(this.width, this.height);
    this.addChildToBack(this._backSprite);
    var c1 = this.bgcolor;
    var c2 = "rgba(0, 0, 0, 0)";
    switch(this.location) {
        case 0:
            var width = Window_ChoiceList.prototype.textWidthEx.call(this, this.text) + 72;
            this._backSprite.bitmap.gradientFillRect(0, 0, width, this.height, c1, c2);
            break;
        case 1:case 4:case 7:
            this._backSprite.bitmap.gradientFillRect(0, 0, this.width / 2, this.height, c1, c2);
            break;
        case 2:case 5:case 8:
            this._backSprite.bitmap.gradientFillRect(this.width / 4, 0, this.width / 4, this.height, c2, c1);
            this._backSprite.bitmap.gradientFillRect(this.width / 2, 0, this.width / 4, this.height, c1, c2);
            break;
        case 3:case 6:case 9:
            this._backSprite.bitmap.gradientFillRect(this.width / 2, 0, this.width / 2, this.height, c2, c1);
            break;
    }
    this._backSprite.opacity = 0;
};

Window_MsgPopup.prototype.refresh = function() {
    this.contents.clear();
    switch(this.location) {
        case 0:case 1:case 4:case 7:
            var x = this.textPadding();
            break;
        case 2:case 5:case 8:
            var x = (this.width - Window_ChoiceList.prototype.textWidthEx.call(this, this.text)) / 2;
            break;
        case 3:case 6:case 9:
            var x = this.width - Window_ChoiceList.prototype.textWidthEx.call(this, this.text) - this.textPadding();
            break;
    }
    this.drawTextEx(this.text, x, 0);
};

Window_MsgPopup.prototype.update = function() {
    Window_Base.prototype.update.call(this);
    this.frameCount++;
    if (this.isFadingIn) {
        this._backSprite.opacity = this.frameCount / JHMW.Message.ToggleFrames * JHMW.Message.WindowOpacity;
        this.contentsOpacity = this.frameCount / JHMW.Message.ToggleFrames * 255;
        this.textX = ((1 - this.frameCount * 2 / JHMW.Message.ToggleFrames) * 100).clamp(0, 100);
        if (this.frameCount >= JHMW.Message.ToggleFrames) {
            this.isFadingIn = false;
            this.isShowing = true;
            this.frameCount = 0;
        }
    } else if (this.isShowing) {
        if (this.frameCount >= JHMW.Message.DisplayFrames) {
            this.isShowing = false;
            this.isFadingOut = true;
            this.frameCount = 0;
        }
    } else if (this.isFadingOut) {
        this._backSprite.opacity = (1 - this.frameCount / JHMW.Message.ToggleFrames) * JHMW.Message.WindowOpacity;
        this.contentsOpacity = (1 - this.frameCount / JHMW.Message.ToggleFrames) * 255;
        if (this.frameCount == JHMW.Message.ToggleFrames) {
            Notification.remove(this.location);
        }
    }
};