//=============================================================================
// Equip Book Plugin
//=============================================================================

/*:
 * @plugindesc A plugin to keep track of obtained equipment and display them in the menu. 
 * @author Your Name
 *
 * @help This plugin does not provide any plugin commands.
 *
 * To add a new equipment to the equip book during the game, use the following code:
 * $gameParty.gainItem($dataArmors[equipmentId], 1);
 * EquipBook.addEquipment(equipmentId);
 *
 * To display the equipped book in a menu, use the following code in a plugin command:
 * SceneManager.push(Scene_EquipBook);
 *
 */

//=============================================================================
// Equipment Book Object
//=============================================================================

function EquipmentBook() {
    this.equipments = {}; // equipment list
}

// add an equipment to the equipment book
EquipmentBook.prototype.addEquipment = function(equipmentId) {
    if (!this.equipments[equipmentId]) {
        this.equipments[equipmentId] = true;
    }
};

// check if an equipment is obtained
EquipmentBook.prototype.isEquipmentObtained = function(equipmentId) {
    return !!this.equipments[equipmentId];
};

// list all obtained equipment
EquipmentBook.prototype.listObtainedEquipments = function() {
    var items = [];
    for (var id in this.equipments) {
        if (this.equipments[id]) {
            var item = $dataArmors[id];
            items.push({ name: item.name, iconIndex: item.iconIndex });
        }
    }
    return items;
};

//=============================================================================
// Equip Book Scene
//=============================================================================

function Scene_EquipBook() {
    this.initialize.apply(this, arguments);
}

Scene_EquipBook.prototype = Object.create(Scene_MenuBase.prototype);
Scene_EquipBook.prototype.constructor = Scene_EquipBook;

Scene_EquipBook.prototype.initialize = function() {
    Scene_MenuBase.prototype.initialize.call(this);
};

Scene_EquipBook.prototype.create = function() {
    Scene_MenuBase.prototype.create.call(this);
    this.createEquipListWindow();
    this.createEquipDetailWindow();
    this.refreshEquipListWindow();
};

Scene_EquipBook.prototype.createEquipListWindow = function() {
    var rect = this.equipListWindowRect();
    this._equipListWindow = new Window_EquipList(rect);
    this._equipListWindow.setHandler('cancel', this.popScene.bind(this));
    this._equipListWindow.setEquipBook($gameParty.equipBook);
    this._equipListWindow.setHandler('equipdetails', this.equipDetail.bind(this));
    this.addWindow(this._equipListWindow);
};

Scene_EquipBook.prototype.createEquipDetailWindow = function() {
    var rect = this.equipDetailWindowRect();
    this._equipDetailWindow = new Window_EquipDetail(rect);
    this._equipListWindow.setStatusWindow(this._equipDetailWindow);
    this.addWindow(this._equipDetailWindow);
};

Scene_EquipBook.prototype.equipListWindowRect = function() {
    var ww = Graphics.boxWidth;
    var wh = Graphics.boxHeight - this._helpWindow.height;
    var wx = 0;
    var wy = this._helpWindow.height;
    return new Rectangle(wx, wy, ww, wh);
};

Scene_EquipBook.prototype.equipDetailWindowRect = function() {
    var ww = Graphics.boxWidth;
    var wh = Graphics.boxHeight - this._equipListWindow.height;
    var wx = 0;
    var wy = this._equipListWindow.height;
    return new Rectangle(wx, wy, ww, wh);
};

Scene_EquipBook.prototype.refreshEquipListWindow = function() {
    this._equipListWindow.refresh();
};

Scene_EquipBook.prototype.equipDetail = function() {
    var equipment = this._equipListWindow.equippedItem();
    this._equipDetailWindow.setEquipment(equipment);
};

//=============================================================================
// Equip List Window
//=============================================================================

function Window_EquipList() {
    this.initialize.apply(this, arguments);
}

Window_EquipList.prototype = Object.create(Window_Selectable.prototype);
Window_EquipList.prototype.constructor = Window_EquipList;

Window_EquipList.prototype.initialize = function(rect) {
    Window_Selectable.prototype.initialize.call(this, rect);
    this._equipBook = null;
    this._data = [];
};

Window_EquipList.prototype.setEquipBook = function(equipBook) {
    this._equipBook = equipBook;
};

Window_EquipList.prototype.maxItems = function() {
    return this._data ? this._data.length : 1;
};

Window_EquipList.prototype.item = function() {
    return this._data && this.index() >= 0 ? this._data[this.index()] : null;
};

Window_EquipList.prototype.equippedItem = function() {
    return this._data && this.index() >= 0 ? $dataArmors[this._data[this.index()].id] : null;
};

Window_EquipList.prototype.isCurrentItemEnabled = function() {
    return this.isEnabled(this._data[this.index()]);
};

Window_EquipList.prototype.setStatusWindow = function(statusWindow) {
    this._statusWindow = statusWindow;
    this.updateStatus();
};

Window_EquipList.prototype.update = function() {
    Window_Selectable.prototype.update.call(this);
    this.updateStatus();
};

Window_EquipList.prototype.updateStatus = function() {
    if (this._statusWindow) {
        this._statusWindow.setEquipment(this.equippedItem());
    }
};

Window_EquipList.prototype.isEnabled = function(item) {
    return true;
};

Window_EquipList.prototype.makeItemList = function() {
    this._data = [];
    for (var i = 1; i < $dataArmors.length; i++) {
        var equip = $dataArmors[i];
        if (equip && equip.etypeId === 1 && equip.name) {
            this._data.push({
                id: equip.id,
                name: equip.name,
                iconIndex: equip.iconIndex,
                obtained: this._equipBook.isEquipmentObtained(equip.id)
            });
        }
    }
};

Window_EquipList.prototype.drawItem = function(index) {
    var item = this._data[index];
    var rect = this.itemRect(index);
    this.changePaintOpacity(item.obtained);
    this.drawIcon(item.iconIndex, rect.x + 2, rect.y + 2);
    this.drawText(item.name, rect.x + 32, rect.y, rect.width - 32);
    this.changePaintOpacity(true);
};

Window_EquipList.prototype.playOkSound = function() {
    if (this.item()) {
        SoundManager.playEquip();
    } else {
        SoundManager.playBuzzer();
    }
};

Window_EquipList.prototype.processOk = function() {
    var item = this.item();
    if (item && item.obtained) {
        this.callHandler('equipdetails');
    } else {
        this.playBuzzerSound();
    }
};

Window_EquipList.prototype.refresh = function() {
    this.makeItemList();
    this.createContents();
    this.drawAllItems();
};

//=============================================================================
// Equip Detail Window
//=============================================================================

function Window_EquipDetail() {
    this.initialize.apply(this, arguments);
}

Window_EquipDetail.prototype = Object.create(Window_Base.prototype);
Window_EquipDetail.prototype.constructor = Window_EquipDetail;

Window_EquipDetail.prototype.initialize = function(rect) {
    Window_Base.prototype.initialize.call(this, rect);
    this._equipment = null;
    this.refresh();
};

Window_EquipDetail.prototype.setEquipment = function(equipment) {
    if (this._equipment !== equipment) {
        this._equipment = equipment;
        this.refresh();
    }
};

Window_EquipDetail.prototype.refresh = function() {
    this.contents.clear();
    if (this._equipment) {
        var y = 0;
        this.drawEquipmentName();
    }
};

Window_EquipDetail.prototype.drawEquipmentName = function() {
    this.changeTextColor(this.systemColor());
    this.drawText(TextManager.equip, 4, 0, 150);
    this.resetTextColor();
    this.drawText(this._equipment.name, 150, 0);
};

//=============================================================================
// Game Party
//=============================================================================

Game_Party.prototype.initialize = function() {
    this._gold = 0;
    this._items = {};
    this._weapons = {};
    this._armors = {};
    this._equipBook = new EquipmentBook();
};

Object.defineProperty(Game_Party.prototype, 'equipBook', {
    get: function() {
        return this._equipBook;
    },
    enumerable: false,
    configurable: true
});

//=============================================================================
// End of File
//=============================================================================
